/*
		JAS v1.00							  (c) 1997 NikSoft International
		--------------------------------------------------------------------
		GLOBALS.H

		Inter-module include file.
		Definition of constants and prototypes imported by all modules.
		Part of this code is copied from the NASM v0.90 Netwide Assembler
		source code.

		09/11/96 09.20	Version v1.00
*/

#ifndef __GLOBALS_H
#define __GLOBALS_H

#define JAS_VER				        "1.3beta"

#define TRUE						1
#define FALSE						0

#define NO_SEG						-1L			/* null segment value */
#define UNDEF_SEG					-2L			/* undefined segment used
												   for extern and global
												   sym defs */
#define SEG_ABS						0x40000000L /* mask for far-absolute
												   segments */
#define MAX_FILENAME				256
#define MAX_STRING					256

/* Error handling constants */
#define ERR_WARNING					0 /* warn only: no further action */
#define ERR_NONFATAL				1 /* terminate assembly after phase */
#define ERR_FATAL					2 /* instantly fatal: exit with error */
#define ERR_PANIC					3 /* internal error: panic instantly
										 and dump core for reference */
#define ERR_MASK					0x03 /* mask off the above codes */

#define ERR_NOFILE					4 /* Don't show file name and lno info */
#define ERR_USAGE					8 /* Show a usage messagge */

/* Error numbers definition */
enum {
  GENERR= 0, ENDERR, IOERR, OPTIONERR, OPTPARAMERR, FMTERR, MEMERR, EXTERR,
  PARSE_ERR, INCLERR, CNTXTERR, INVPARAM, EXTRAPARAMERR, PARAMNOERR,
  UNDERR, LBLERR, INSERR, SEGERR, INVPERR, INVSECERR, TYPERR, DIVERR,
  OPRERR, INTERR, SIZERR, INVSIZERR, NOPROCERR, EXPERR, OPRSERR, PREFSERR,
  EAERR, ASSERR, INVDIREC, UNDEFSYMSERR, EXITERR, TBLERR, MAXERROR
};

/* size, and other attributes, of the operand */
#define BITS8						0x00000001L
#define BITS16						0x00000002L
#define BITS32						0x00000004L
#define BITS64						0x00000008L /* FPU only */
#define BITS80						0x00000010L /* FPU only */
#define FAR							0x00000020L /* grotty: this means 16:16
												   or 16:32, like in
												   CALL/JMP */
#define NEAR						0x00000040L
#define SHORT						0x00000080L

#define SIZE_MASK					0x000000FFL /* all the size attributes */
#define NON_SIZE					(~SIZE_MASK)

#define TO							0x00000100L /* reverse effect in FADD,
												   FSUB &c */
#define COLON						0x00000200L /* operand is followed by
												   a colon */

/* type of operand: memory reference, register, etc. */
#define MEMORY						0x00204000L
#define REGISTER					0x00001000L /* register number in
												   'basereg' */
#define IMMEDIATE					0x00002000L

#define REGMEM						0x00200000L /* for r/m, ie EA, operands */
#define REGNORM						0x00201000L /* 'normal' reg, qualifies
												   as EA */
#define REG8						0x00201001L
#define REG16						0x00201002L
#define REG32						0x00201004L
#define FPUREG						0x01000000L /* floating point stack
												   registers */
#define FPU0						0x01000800L /* FPU stack register zero */

/* special register operands: these may be treated differently */
#define REG_SMASK					0x00070000L /* a mask for the following */
#define REG_ACCUM					0x00211000L /* accumulator: AL, AX or
												   EAX */
#define REG_AL						0x00211001L /* REG_ACCUM | BITSxx */
#define REG_AX						0x00211002L /* ditto */
#define REG_EAX						0x00211004L /* and again */
#define REG_COUNT					0x00221000L /* counter: CL, CX or ECX */
#define REG_CL						0x00221001L /* REG_COUNT | BITSxx */
#define REG_CX						0x00221002L /* ditto */
#define REG_ECX						0x00221004L /* another one */
#define REG_DX						0x00241002L
#define REG_SREG					0x00081002L /* any segment register */
#define REG_CS						0x01081002L /* CS */
#define REG_DESS					0x02081002L /* DS, ES, SS (non-CS 86
												   registers) */
#define REG_FSGS					0x04081002L /* FS, GS (386 extended
												   registers) */
#define REG_CDT						0x00101004L /* CRn, DRn and TRn */
#define REG_CREG					0x08101004L /* CRn */
#define REG_CR4						0x08101404L /* CR4 (Pentium only) */
#define REG_DREG					0x10101004L /* DRn */
#define REG_TREG					0x20101004L /* TRn */

/* special type of EA */
#define MEM_OFFS					0x00604000L /* simple [address] offset */

/* special type of immediate operand */
#define UNITY						0x00802000L /* for shift/rotate
												   instructions */

/* This is a useful #define which I keep meaning to use more often:
   the number of elements of a statically defined array.  */

#define elements(x)		( sizeof(x) / sizeof(*(x)) )

/* ------------------------------------------------------------- */

/* Max number of allowed errors. */
extern short	MaxErrors;

/* Number of errors encountered so far. */
extern short	Errors;

/* Controls case sensitivity of the identifiers.
   0	 = Don't care about case, all identifier are converted to upper case
   non 0 = Identifiers are case sensitive ( dummy != Dummy ).
		   This is the default. */
extern short	IdCase;

/* Defined and mantained by the lexer */
extern unsigned int yylineno;

/* Error reporting function */
extern void error(int, int, ...);

/* Returns the name of the input/output file. */
extern char *getInputFileName();
extern char *getOutputFileName();

#endif

/* --------- End of File --------------------------------------- */
