/*
		JAS v1.00							  (c) 1996 NikSoft International
		--------------------------------------------------------------------
		PARSER.Y

		Bison grammar specification for the parser.
		The syntax of the used grammar is TASM-like, with various directive
		typical of each output file format generated by the assembler.

		03/11/96 18.02	Version 1.00
		Friday, February 07, 1997 Version 1.01
			Improved syntax and semantic actions for operand parsing.
		Thursday, February 20, 1997
			The original project is now called JAS The Jet ASsembler,
			simply because I like more this name!
			All the function and datas concerning expressions evaluation
			have been moved in a stand alone module.
		Sunday, May 11, 1997
			Added backpatch and SymRefs() code for the data definition
			backpatch.
		Tuesday, July 22, 1997
			Added #define style macro definition.
		Sunday, August 03, 1997
			Improved macro and macro processing, added the processing
			of the .REPEAT directive.
		Thursday, August 07, 1997
			Added processing of the .MACRO directive.
		Saturday, November 22, 1997
			Added boolean expressions parsing.
			Added conditional assembling directives.
*/

%{
#include <stdarg.h>

#include "globals.h"
#include "jaslib.h"

#include "ptypes.h"
#include "lexan.h"
#include "symmngr.h"
#include "bpatch.h"

#include "assemble.h"
#include "eval.h"

#define PARAM_DELTA				8
%}

%union {
	EXPR_RESULT		*result;
	OPERAND			instrOp;
	struct symbol	*identifier;
	long			intval;
	char			*strval;
	char			chrval;
	float			floatval;
	struct {
		int DeclType;
		int DeclSize;
	}				vardecl;

	struct {
		int		Lines;
		char	*text;
	}				textval;
}

%token			<identifier>	TOKEN_ID
%token			<intval>		TOKEN_NUM
%token			<intval>		TOKEN_COUNT
%token			<strval>		TOKEN_STR
%token			<strval>		TOKEN_DIRECTIVE
%token			<textval>		TOKEN_TEXT
%token			<chrval>		TOKEN_CHR
%token			<floatval>		TOKEN_FLT

%token			<intval>		TOKEN_REG
%token			<intval>		TOKEN_INSN
%token			<intval>		TOKEN_DDATA

%token			<intval>		TOKEN_PREFIX
%token			<intval>		TOKEN_SPECIAL

%token			<intval>		TOKEN_EQ
%token			<intval>		TOKEN_CMP

%token			TOKEN_ALLD
%token			TOKEN_ALLW

%token			TOKEN_SECTION
%token			TOKEN_BITS
%token			TOKEN_INCLUDE
%token			TOKEN_INCBIN
%token			TOKEN_ALIGN
%token			TOKEN_COMMON
%token			TOKEN_GLOBAL
%token			TOKEN_EXTERN
%token			TOKEN_DEFINE
%token			TOKEN_REPEAT
%token			TOKEN_MACRO
%token			TOKEN_END
%token			TOKEN_IF
%token			TOKEN_ELSE

%token			TOKEN_BASE
%token			TOKEN_HERE

%token			TOKEN_SHL
%token			TOKEN_SHR
%token			TOKEN_DIV
%token			TOKEN_MOD
%token			TOKEN_AND
%token			TOKEN_OR
%token			TOKEN_XOR
%token			TOKEN_NOT

%token			TOKEN_DEF

%token			TOKEN_PROC
%token			TOKEN_USES
%token			TOKEN_ARG
%token			TOKEN_LOCAL
%token			TOKEN_ENDP
%token			TOKEN_TIMES

/* --------- Non terminals type definition --------------------- */

%type			<result>		expr
%type			<result>		expr0
%type			<result>		part_2

%type			<identifier>	label

%type			<intval>		register
%type			<intval>		parameters
%type			<intval>		paramdecl
%type			<intval>		localdecl
%type			<intval>		localvar
%type			<intval>		locals
%type			<intval>		count
%type			<intval>		sizovr
%type			<intval>		size_overrides
%type			<intval>		addr_override
%type			<intval>		incbin_param
%type			<intval>		bexpr
%type			<intval>		bexpr0

%type			<instrOp>		imm_or_mem
%type			<instrOp>		operand_spec

%type			<vardecl>		typespec

/* --------- Operators precedence and associativity ------------ */

%left		TOKEN_EQ
%left		TOKEN_CMP
%left		TOKEN_XOR
%left		TOKEN_OR
%left		TOKEN_AND
%left		TOKEN_SHL TOKEN_SHR
%left		'+' '-'
%left		'*' '/' TOKEN_DIV TOKEN_MOD
%left		UNARY_MINUS TOKEN_NOT

/* --------- Data local to the parser --------------------------- */

%{
/* The instruction being parsed and assembled */
static INSN instr;

/* sizes and special flags */
static long reg_flags[] = {
	0, REG8, REG_AL, REG_AX, REG8, REG8, REG16, REG16, REG8, REG_CL,
	REG_CREG, REG_CREG, REG_CREG, REG_CR4, REG_CS, REG_CX, REG8,
	REG16, REG8, REG_DREG, REG_DREG, REG_DREG, REG_DREG, REG_DREG,
	REG_DREG, REG_DESS, REG_DX, REG_EAX, REG32, REG32, REG_ECX,
	REG32, REG32, REG_DESS, REG32, REG32, REG_FSGS, REG_FSGS, REG16,
	REG16, REG_DESS, FPU0, FPUREG, FPUREG, FPUREG, FPUREG, FPUREG,
	FPUREG, FPUREG, REG_TREG, REG_TREG, REG_TREG, REG_TREG, REG_TREG
};

static struct ProcInfo {
	SYMBOL	*Id;
	int		RegList[20];
	short	RegCount;
	int		LocalSize;
	int		ParamSize;
} CurrProc;

static SYMBOL	*SecId;
static long		Segment;
static long		Offset;
static long		LocalOffset;

static short	OprsCount;
static OPERAND	*TempOp;

static short	mref;
static PATCH	*PatchPtr;
static EXTOP	**eoptail;

static long		anInt;
static float	aFloat;

static int		ParamPos;
static int		ParamSize;
static char		**ParamPtr;
%}

/* -------- Prototype of function local to the parser ----------------- */
%{

/* Parser's utility functions */
static long		AssembleProcProlog(struct ProcInfo *);
static long		AssembleProcEpilog(struct ProcInfo *);
static long		IncludeBinary(char *, long, long);
static int		DefineLabel(SYMBOL *);
static OPERAND	*CalcMemRef(register EXPR_RESULT *, register EXPR_RESULT *);
static OPERAND	*ProcessOperand(EXPR_RESULT *);
static EXTOP	*ProcessExtop(int, EXPR_RESULT *);
static int		ProcessOverride(int);
static void		InstrCleanup();
static int		SymRefs(register EXPR_RESULT *, short, ...);
static int		yyerror(char *);
%}

/* --------- Public datas -------------------------------------- */

%{
/* Additional token info */
int TokenInfo;
short int CodeBits;
%}

%%

lines	: lines line
	| line
	;

line : aline
		{
			yylineno++;
			SetExpr(0);
			EvaluatorReset();
		}
	;


/*
	The .REPEAT directive is treated as a line and not as
	a directive, because the parser will reduce the production on the '\n'
	token and not on some other token causing syntax errors due to the
	lexical context switch. */
aline	:	dummy '\n'

	|	TOKEN_REPEAT expr '\n' TOKEN_TEXT TOKEN_END '\n'
		{
			SetCritical(-1);
			if ( IsSimple($2) != INTEGER ) {
				error(ERR_NONFATAL, INTERR);
				YYERROR;
			};
			yylineno++;
			RepeatText($4.text, $4.Lines, IntVal($2));
			/* The text stored has been copied in the new scanning
			   buffer, then we can free it up. */
			LIB_free($4.text);
		}

	|	TOKEN_IF bexpr '\n' TOKEN_TEXT TOKEN_END '\n'
		{
			yylineno++;
			if ( $2 ) {
				RepeatText($4.text, $4.Lines, 1);
			};

			LIB_free($4.text);
		}

	|	TOKEN_IF bexpr '\n' TOKEN_TEXT TOKEN_ELSE '\n' TOKEN_TEXT TOKEN_END '\n'
		{
			yylineno += 2;
			if ( $2 != FALSE ) {
				RepeatText($4.text, $4.Lines, 1);
			} else {
				RepeatText($7.text, $7.Lines, 1);
			};

			LIB_free($4.text);
			LIB_free($7.text);
		}

	|	TOKEN_INCLUDE TOKEN_STR '\n'
		{   SwitchToFile($2);   }

	| error '\n'
	    {
			InstrCleanup();
			yyerrok;
		}
	;

dummy	:	directive {}

	|	label prefixes assembler_instr
		{
		    /* Clear the instr struct for the next instruction */
			if ( instr.opcode != -1 ) InstrCleanup();
		}

	|	TOKEN_ID TOKEN_PROC
		{
			if ( CurrProc.Id ) {
				error(ERR_NONFATAL, GENERR,
					"Nested procedures not implemented yet!");
				FreeSymbol($1);
				YYERROR;
			};

			if ( $1->Type != NO_TYPE && IDType($1) != LABEL ) {
				error(ERR_NONFATAL, LBLERR, $1->Name);
				YYERROR;
			};

			/* This is used to link the local symbols (parameters
			   and variables) to the defined procedure. */
			CurrProc.Id= $1;
		}
		uses parameters locals
		{
			/* Define the label for the procedure */
			if ( DefineLabel($1) ) YYERROR;

			/* Fill in the 'CurrProc' struct. */
			CurrProc.ParamSize= $5;
			CurrProc.LocalSize= $6;

			Offset += AssembleProcProlog(&CurrProc);
		}

	|	TOKEN_ID TOKEN_ENDP
		{
			FreeSymbolList($1);

			if ( $1 != CurrProc.Id ) {
				error(ERR_NONFATAL, UNDERR, $1->Name);
				FreeSymbol($1);
			};

			/* Clear the procedure info struct. */
			bzero(&CurrProc, sizeof(struct ProcInfo));
		}

	|	TOKEN_ENDP
		{
			if ( CurrProc.Id == NULL ) {
				error(ERR_NONFATAL, NOPROCERR);
			} else {
				/* Undefine all symbols linked to the last defined
				   procedure. */
				FreeSymbolList( CurrProc.Id );

			    /* Clear the procedure info struct. */
			    bzero(&CurrProc, sizeof(struct ProcInfo));
			};
		}
	;

assembler_instr :	/* Empty */		{ instr.opcode = -1; }
	|	TOKEN_INSN
		{
			instr.opcode	= $1;
			instr.condition = TokenInfo;
		}
		operands
		{
			switch ( $1 ) {
				case I_RESW:
					instr.opcode= I_RESB;
					instr.oprs[0].offset *= 2;
					break;

				case I_RESD:
					instr.opcode= I_RESB;
					instr.oprs[0].offset *= 4;
					break;

				case I_RESQ:
					instr.opcode= I_RESB;
					instr.oprs[0].offset *= 2;
					break;

				case I_REST:
					instr.opcode= I_RESB;
					instr.oprs[0].offset *= 10;
					break;

				case I_RET:
				case I_RETF:
				case I_RETN:
				case I_IRETW:
				case I_IRETD:
				case I_IRET:
					/* This is a return instruction assemble code
					   for the procedure epilog */
					Offset += AssembleProcEpilog( &CurrProc );
			};

			/* Fill in the 'patch_instr' field of the patch.
			   We must do it here because the all parsing for
			   the istruction is done and the 'instr' struct is
			   filled properly. */
			if ( PatchPtr ) {
				PatchPtr->patch_instr= instr;
				/* Calculate the size of the istruction and
				   pad the code with NOP operations */
				Offset += pad(Segment, Offset, '\x90',
				    instrsize(Segment, Offset, CodeBits, &instr));
			} else {
				Offset += assemble(Segment, Offset, CodeBits, &instr);
			}
		}

	|	TOKEN_DDATA data_list
		{
			/* Fill in some instr field */
			instr.opcode	= $1;
			instr.condition = -1;

			if ( instr.nprefix ) {
				instr.nprefix = 0;
				error(ERR_WARNING, GENERR, "Prefixes ignored");
			};

			Offset += assemble(Segment, Offset, CodeBits, &instr);
		}
	;

locals	:	/* Empty */ { $$= 0; }
	|	TOKEN_LOCAL
		{
			LocalOffset= ( CodeBits == 16 ? -2 : -4 );
			EvaluatorReset();
		}
		localvar
		{
			$$ = $3;
		}
	;

localvar	:	localdecl
		{
			LocalOffset -= $1;
			$$ = $1;
		}

	|	localvar ',' localdecl
		{
			LocalOffset -= $3;
			$$ = $1 + $3;
		}
	;

uses	:
	|	registers
	;

registers	:	TOKEN_USES	{ CurrProc.RegCount = 0 }
	register
		{
			CurrProc.RegList[CurrProc.RegCount++]= $3;
		}
	|	registers ',' register
		{
			if ( CurrProc.RegCount == elements(CurrProc.RegList) ) {
			    /* Don't save last register */
				error(ERR_WARNING, GENERR,
					"Register not saved, too much registers "
					"(try USES ALL).");
			} else {
			    CurrProc.RegList[CurrProc.RegCount++]= $3;
			};
		}
	;

register	:	TOKEN_REG	{ $$ = $1; }
	|	TOKEN_ALLD
		{	/* Uses all long registers */
			$$ = REG_ENUM_LIMIT;
		}
	|	TOKEN_ALLW
		{	/* Uses all word registers */
			$$ = REG_ENUM_LIMIT + 1;
		}
	;

parameters	:       { $$=  0; }
	|	paramdecl   { $$= $1; }
	;

paramdecl	:	TOKEN_ARG
		{
			LocalOffset= ( CodeBits == 16 ? 4 : 8 );
			EvaluatorReset();
		}
	localdecl
		{
			LocalOffset += $3;
			$$ = $3;
		}
	|	paramdecl ',' localdecl
		{
			LocalOffset += $3;
			$$ = $1 + $3;
		}
	;

localdecl	:	TOKEN_ID ':' typespec
	    {
			if ( $1->Type != NO_TYPE ) {
				error(ERR_NONFATAL, LBLERR, $1->Name);
				$$ = 0;
			} else {
				LinkSymbol(CurrProc.Id, $1);
				$1->Type = (LABEL | LOCAL) | $3.DeclType;
				$1->Value= LocalOffset;
				$$= $3.DeclSize;
			};

			EvaluatorReset();
		}
	;

typespec	:	TOKEN_SPECIAL count
		{
			switch ( $1 ) {
				case S_WORD:
					$$.DeclType = BITS16;
					$$.DeclSize = 2;
					break;

				case S_DWORD:
					$$.DeclType = BITS32;
					$$.DeclSize = 4;
					break;

				case S_QWORD:
					$$.DeclType = BITS64;
					$$.DeclSize = 8;
					break;

				case S_TWORD:
					$$.DeclType = BITS80;
					$$.DeclSize = 10;
					break;

				case S_FAR:
					$$.DeclType = FAR;
					$$.DeclSize = ( CodeBits == 16 ? 4 : 6);
					break;

				case S_NEAR:
					$$.DeclType = NEAR;
					$$.DeclSize = ( CodeBits == 16 ? 2 : 4);
					break;

				default:
					error(ERR_NONFATAL, SIZERR);

				case S_BYTE:
					$$.DeclType = BITS8;
					$$.DeclSize = 1;
			};

			$$.DeclSize *= $2;
		}
	;


count	:           { $$ = 1; }
	|	':' { SetExpr(-1); } expr
		{
			SetCritical(-1);
			if ( IsSimple($3) != INTEGER ) {
				error(ERR_NONFATAL, INTERR);
				$$ = 1;
			} else {
			    $$= $3->int_result;
			};

			SetCritical(0);
			SetExpr(0);
		}
	;

/* --------- Directive processing ------------------------------ */

directive	:	TOKEN_DIRECTIVE
	    {
			/* Handle format specific directive */
			if ( Directive($1) ) YYERROR;
		}

	|	TOKEN_SECTION TOKEN_ID
		{
			anInt= Section($2->Name);

			if ( anInt == NO_SEG ) {
			/* Section name is not valid */
				error(ERR_NONFATAL, INVSECERR, $2);
				YYERROR;
			};

			/* Define some identifier fields if not defined */
			if ( $2->Type == NO_TYPE ||
				($2->Type == (SECLBL | LABEL) && $2->Value == -1) ) {
				$2->Type   = LABEL | SECLBL;
				$2->segment= anInt;
				$2->offset = Offset;
				$2->Value  = 0;

				/* Do the backpatch */
				BPatch($2);
			} else if ( $2->Type != (LABEL | SECLBL) ) {
				error(ERR_NONFATAL, LBLERR, $2->Name);
				YYERROR;
			};

			/* Save the offset reached in the current section */
			SecId->Value= Offset;

			/* Destroy all the local symbol defined in the
			   current section */
			FreeSymbolList(SecId);

			/* An now change the section ...
			   Restore the old offset and switch to the
			   segment of the new section */
			Offset = $2->Value;
			Segment= anInt;

			SecId= $2;
		}

	|	TOKEN_BITS TOKEN_NUM
		{
			CodeBits= $2;

			if ( CodeBits != 16 && CodeBits != 32 ) {
				error(ERR_WARNING, GENERR,
					"Defaulting to 16 bits assembling");
				CodeBits = 16;
			};
		}

	|	TOKEN_INCBIN TOKEN_STR incbin_param
		{ Offset += IncludeBinary($2, $3, anInt); }

	|	TOKEN_ALIGN expr
		{
			SetCritical(-1);
			if ( IsSimple($2) != INTEGER ) {
				error(ERR_NONFATAL, INTERR);
				YYERROR;
			};

			if ( (anInt= IntVal($2)) ) {
				if ( (Offset % anInt) ) {
					Offset +=
						pad(Segment, Offset, '\x90', anInt - (Offset % anInt));
				};
			};

			EvaluatorReset();
		}

	|	TOKEN_COMMON TOKEN_ID ':' typespec
		{
			if ( $2->Type != NO_TYPE ) {
				error(ERR_NONFATAL, LBLERR, $2->Name);
				YYERROR;
			};

			/* Defines the symbol's fields. */
			$2->Type   = (LABEL | COMMLBL) | $4.DeclType;
			$2->offset = 0;
			$2->segment= seg_alloc();
			$2->Value  = $4.DeclSize;

			/* Defines the symbol in the output format */
			Symdef($2, 2);

			/* Symbol backpatch. */
			BPatch($2);
		}

	|	TOKEN_GLOBAL TOKEN_ID
		{
			if ( $2->Type == NO_TYPE ||
				 ( IDType($2) == LABEL && IDFlag($2) == GLBLBL) ) {
			/* Defines the symbol's fields. */

				if ( $2->Type == NO_TYPE ) {
				/* The symbol has not been defined yet, so
				   declare it as a global symbol */
					$2->Type   = LABEL | GLBLBL;
					$2->offset = 0;
					$2->segment= UNDEF_SEG;
					$2->Value  = 0;
				} else {
				/* The symbol has been defined by the assembled
				   source file, define it in the output
				   format */
					$2->Type   = LABEL | GLBLBL;
					Symdef($2, 1);
				}
			} else {
				error(ERR_NONFATAL, LBLERR, $2->Name);
				YYERROR;
			};

		}

	|	TOKEN_EXTERN TOKEN_ID
		{
			if ( $2->Type != NO_TYPE ) {
				error(ERR_NONFATAL, LBLERR, $2->Name);
				YYERROR;
			};

			/* Defines the symbol's fields. */
			$2->Type   = LABEL | EXTLBL;
			$2->offset = 0;
			$2->segment= seg_alloc();
			$2->Value  = 0;

			/* --- Define the symbol in the output format */
			Symdef($2, 2);
		}

	|	TOKEN_DEFINE TOKEN_ID TOKEN_STR
		{
			if ( $2->Type != NO_TYPE ) {
				error(ERR_NONFATAL, LBLERR, $2->Name);
				YYERROR;
			};

			/* Defines the symbol's fields. */
			$2->Type     = MACRO | DEFINED;
			$2->offset   = -1;
			$2->segment  = -1;
			$2->Value    = 0;
			$2->ExtraLong= 0;

			BuildMacro($2, NULL, $3, 0);
		}

	|	TOKEN_MACRO TOKEN_ID macro_param_def '\n' TOKEN_TEXT TOKEN_END
		{
			if ( $2->Type != NO_TYPE ) {
				error(ERR_NONFATAL, LBLERR, $2->Name);
				YYERROR;
			};

			/* Defines the symbol's fields. */
			$2->Type     = MACRO;
			$2->offset   = -1;
			$2->segment  = -1;
			$2->Value    = 0;
			$2->ExtraLong= ParamPos;

			/* Build up the macro. */
			BuildMacro($2, ParamPtr, $5.text, $5.Lines);
			ParamPtr = NULL;
			ParamPos = 0;
			ParamSize= 0;
		}

	;

macro_param_def :	/* Empty */ { ParamPtr = NULL; ParamPos= 0; }
	| macro_params
		{
			ParamPtr[ParamPos]= NULL;
		}
	;

macro_params	:	macro_params ',' macro_param
	|	macro_param
	;

macro_param		:	TOKEN_STR
		{
			if ( ParamPos == ParamSize ) {
			/* Allocates space for the parameters array. */
				ParamSize += PARAM_DELTA;
				ParamPtr= (char **)LIB_realloc(ParamPtr,
					sizeof(char *) * (ParamSize + 1));
			};

			ParamPtr[ParamPos++]= $1;
		}
	;

incbin_param :	/* Empty */ { $$ = -1; anInt= 0; }
	|	',' TOKEN_NUM       { $$ = $2; anInt= 0; }
	|	',' TOKEN_NUM ',' TOKEN_NUM { $$= $2; anInt= $4; }

/* --------- Label declaration --------------------------------- */

label	:	/* Empty */		{ $$ = NULL; }
	|	TOKEN_ID ':'
		{
			/* See if the label has been already defined */
			if ( $1->Type != NO_TYPE && IDType($1) != LABEL ) {
				error(ERR_NONFATAL, LBLERR, $1->Name);
				YYERROR;
			};

			if ( DefineLabel($1) ) YYERROR;
			$$= $1;
		}
	;

/* --------- Prefixes processing ------------------------------- */

prefixes	:	/* Empty */ { instr.nprefix = 0; }
	|	prefs
	;

prefs	:	prefix
	|	prefs prefix
	;

prefix	:	TOKEN_REG
		{
			if ( instr.nprefix == 3 ) {
				error(ERR_FATAL, PREFSERR);
				YYERROR;
			};

			if ( REG_SREG & ~reg_flags[$1] ) {
				error(ERR_FATAL, SEGERR);
				YYERROR;
			};

			instr.prefixes[instr.nprefix++]= $1;
		}

	|	TOKEN_PREFIX
		{
			if ( instr.nprefix == 3 ) {
				error( ERR_FATAL, PREFSERR);
				YYERROR;
			};

			instr.prefixes[instr.nprefix++]= $1;
		}

	|	TOKEN_TIMES { SetExpr(-1); } expr
		{
			SetCritical(-1);
			SetExpr(0);
			if ( IsSimple($3) != INTEGER ) {
				error(ERR_NONFATAL, INTERR);
				YYERROR;
			};
			instr.times = IntVal($3);
			EvaluatorReset();
		}
	;

/* ------------------------------------------------------------- */

operands	:	/* Empty */
		{
			instr.operands = 0;
			bzero(&instr.oprs, 3*sizeof(OPERAND));
		}
	|	    {
			    if ( $<intval>0 >= I_RESB && $<intval>0 <= I_RESW ) {
			    /* Set the critical flag if we are parsing operands for
			       a RES? istruction. */
				    SetCritical(-1);
			    };
		    }
		operandlist
			{
				instr.operands = OprsCount;
				/* Clear remaining operands */
				while ( OprsCount < 3 ) {
					instr.oprs[OprsCount].type = 0;
					OprsCount++;
				};
			}
	;

operandlist :	operand
	|	operandlist ',' operand
	;

operand :	operand_spec
	    {
		    if ( OprsCount == 3 ) {
			    error(ERR_NONFATAL, OPRSERR);
			    YYERROR;
		    };
			instr.oprs[OprsCount] = $1;
			OprsCount++;
		}

	|	operand_spec ':' operand_spec
		{
			if ( OprsCount > 1 ) {
				error(ERR_NONFATAL, OPRSERR);
				YYERROR;
			};

			instr.oprs[OprsCount]	 = $1;
			instr.oprs[OprsCount + 1]= $3;
			instr.oprs[OprsCount].type |= COLON;

			OprsCount += 2;
		}
	;


operand_spec	:	sizovr imm_or_mem
	    {
			$$ = $2;
			$$.type |= $1;
		}
	;

imm_or_mem	:	expr
		{
			TempOp = ProcessOperand( $1 );
			if ( !TempOp ) {
			/* Error during operand conversion */
				YYERROR;
			};
			$$ = *TempOp;
		}

	|	'[' addr_override expr part_2 ']'
		{
			TempOp = ( $4 ? CalcMemRef($3, $4) : CalcMemRef(NULL, $3) );

			if ( !TempOp ) {
			/* Error during operand conversion */
				YYERROR;
			};

			$$ = *TempOp;
			$$.addr_size = $2;
		}
	;

part_2	:	/* Empty */ { $$ = NULL; }
	|	':' expr        { $$ = $2; }
	;

sizovr	:	/* Empty */ { $$ = 0; }
	|	size_overrides	{ $$ = $1; }
	;

size_overrides	:	TOKEN_SPECIAL
		{
			if ( ($$ = ProcessOverride($1)) == -1 ) {
				YYERROR;
			};
		}
	|	size_overrides TOKEN_SPECIAL
		{
			if ( ($$ = $1 | ProcessOverride($2)) == -1 ) {
				YYERROR;
			};
		}
	;

addr_override	:	/* Empty */ { $$ = 0; }
	|	TOKEN_SPECIAL
		{
			switch ( $1 ) {
				case S_WORD:
					$$ = 2;
					break;

				case S_DWORD:
					$$ = 4;
					break;

				 default:
					error(ERR_NONFATAL, INVSIZERR);
					YYERROR;
			};
		}
	;

/* -------- Data definition ------------------------------------------- */

data_list	:	/* Empty */ { instr.eops= NULL; }
	|	datas		        { *eoptail= NULL; }
	;

datas	:	expr
		{
			if ( !(instr.eops= ProcessExtop($<intval>0, $1)) ) {
				YYERROR;
			};

			eoptail= &(instr.eops->next);
		}
	|	datas ',' expr
		{
			if ( !(*eoptail= ProcessExtop($<intval>0, $3)) ) {
				YYERROR;
			};

			eoptail= &(*eoptail)->next;
		}
	;

/* --------- Expression processing ----------------------------- */

expr	:	expr0			{ $$ = $1; }
	|	'-' expr %prec UNARY_MINUS  { $$ = ScalarMult(-1, $2); }
	|	TOKEN_NOT expr
		{
			SetCritical(-1);

			if ( IsSimple($2) != INTEGER ) {
				error(ERR_NONFATAL, INTERR);
				YYERROR;
			};

			$$= Vectorize(INTEGER, 1, ~IntVal($2));
			SetCritical(0);
		};

	|	'(' expr ')'    { $$ = $2; }
	|	expr '+' expr   { $$ = AddExprs($1, $3); }
	|	expr '-' expr   { $$ = AddExprs( $1, ScalarMult(-1, $3) ); }
	|	expr TOKEN_DIV expr
	    {
			SetCritical(-1);
			if ( (anInt = IsSimple($1)) == -1) {
				error(ERR_NONFATAL, EXPERR);
				YYERROR;
			};

			if ( anInt != IsSimple($3) ) {
				error(ERR_NONFATAL, TYPERR);
				YYERROR;
			};

			if ( anInt == INTEGER ) {
				if ( !(anInt = IntVal($3)) ) {
					error(ERR_NONFATAL, DIVERR);
					YYERROR;
				};

				anInt = IntVal($1) / anInt;
				$$ = Vectorize(INTEGER, 1, anInt);
			} else {
				error(ERR_NONFATAL, OPRERR);
				YYERROR;
			};
			SetCritical(0);
		}

	|	expr TOKEN_MOD expr
		{
			SetCritical(-1);
			if ( (anInt = IsSimple($1)) == -1) {
				error(ERR_NONFATAL, EXPERR);
				YYERROR;
			};

			if ( anInt != IsSimple($3) ) {
				error(ERR_NONFATAL, TYPERR);
				YYERROR;
			};

			if ( anInt == INTEGER ) {
				if ( !(anInt = IntVal($3)) ) {
					error(ERR_NONFATAL, DIVERR);
					YYERROR;
				};

				anInt = IntVal($1) % anInt;
				$$ = Vectorize(INTEGER, 1, anInt);
			} else {
				error(ERR_NONFATAL, OPRERR);
				YYERROR;
			};

			SetCritical(0);
		}

	|	expr '*' expr
		{
			switch ( IsSimple($1) ) {
				case INTEGER:
					switch ( IsSimple($3) ) {
						case INTEGER:
					        $$ = Vectorize(INTEGER, 1,
						            IntVal($1) * IntVal($3) );
					        break;

						case FLOAT:
							$$ = Vectorize(FLOAT, 1,
								(float)(IntVal($1) * FloatVal($3)) );
							break;

						case -1:
							$$ = ScalarMult(IntVal($1), $3);
							break;

						default:
							error(ERR_NONFATAL, TYPERR);
				            YYERROR;
					};
					break;

				case FLOAT:
					switch ( IsSimple($3) ) {
						case INTEGER:
					        $$ = Vectorize(FLOAT, 1,
						            (float)(FloatVal($1) * IntVal($3)) );
					        break;

						case FLOAT:
							$$ = Vectorize(FLOAT, 1,
								FloatVal($1) * FloatVal($3) );
							break;

						default:
							error(ERR_NONFATAL, TYPERR);
				            YYERROR;
					};
					break;

				case -1:
					if ( IsSimple($3) == INTEGER ) {
						$$ = ScalarMult(IntVal($3), $1);
					} else {
						error(ERR_NONFATAL, INTERR);
						YYERROR;
					}
					break;

				default:
					error(ERR_NONFATAL, OPRERR);
					YYERROR;
			};
		}

	|	expr '/' expr
		{
			SetCritical(-1);
			if ( (anInt = IsSimple($1)) == -1) {
				error(ERR_NONFATAL, EXPERR);
				YYERROR;
			};

			if ( anInt != -1 && anInt == IsSimple($3) ) {
				error(ERR_NONFATAL, TYPERR);
				YYERROR;
			};

			switch ( anInt ) {
				case INTEGER:
					if ( !(anInt = IntVal($3)) ) {
						error(ERR_NONFATAL, DIVERR);
						YYERROR;
					};

					anInt = IntVal($1) / anInt;
					$$ = Vectorize(INTEGER, 1, anInt);
					break;

				case FLOAT:
					if ( !(aFloat = FloatVal($3)) ) {
						error(ERR_NONFATAL, DIVERR);
						YYERROR;
					};

					aFloat = FloatVal($1) / aFloat;
					$$ = Vectorize(FLOAT, 1, aFloat);
					break;

				default:
					error(ERR_NONFATAL, OPRERR);
					YYERROR;
			};
			SetCritical(0);
		}

	|	expr TOKEN_SHR expr
		{
			SetCritical(-1);
			if ( (anInt = IsSimple($1)) == -1) {
				error(ERR_NONFATAL, EXPERR);
				YYERROR;
			};

			if ( anInt != IsSimple($3) ) {
				error(ERR_NONFATAL, TYPERR);
				YYERROR;
			};

			if ( anInt != INTEGER ) {
				error(ERR_NONFATAL, OPRERR);
				YYERROR;
			};

			anInt= IntVal($1) >> IntVal($3);
		    $$ = Vectorize(INTEGER, 1,  anInt);
			SetCritical(0);
		}

	|	expr TOKEN_SHL expr
		{
			SetCritical(-1);
			if ( (anInt = IsSimple($1)) == -1) {
				error(ERR_NONFATAL, EXPERR);
				YYERROR;
			};

			if ( anInt != IsSimple($3) ) {
				error(ERR_NONFATAL, TYPERR);
				YYERROR;
			};

			if ( anInt != INTEGER ) {
				error(ERR_NONFATAL, OPRERR);
				YYERROR;
			};

			anInt= IntVal($1) << IntVal($3);
		    $$ = Vectorize(INTEGER, 1,  anInt);
			SetCritical(0);
		}

	|	expr TOKEN_AND expr
		{
			SetCritical(-1);
			if ( (anInt = IsSimple($1)) == -1) {
				error(ERR_NONFATAL, EXPERR);
				YYERROR;
			};

			if ( anInt != IsSimple($3) ) {
				error(ERR_NONFATAL, TYPERR);
				YYERROR;
			};

			if ( anInt != INTEGER ) {
				error(ERR_NONFATAL, OPRERR);
				YYERROR;
			};

			anInt= IntVal($1) & IntVal($3);
		    $$ = Vectorize(INTEGER, 1,  anInt);
			SetCritical(0);
		}

	|	expr TOKEN_OR expr
		{
			SetCritical(-1);
			if ( (anInt = IsSimple($1)) == -1) {
				error(ERR_NONFATAL, EXPERR);
				YYERROR;
			};

			if ( anInt != IsSimple($3) ) {
				error(ERR_NONFATAL, TYPERR);
				YYERROR;
			};

			if ( anInt != INTEGER ) {
				error(ERR_NONFATAL, OPRERR);
				YYERROR;
			};

			anInt= IntVal($1) | IntVal($3);
		    $$ = Vectorize(INTEGER, 1,  anInt);
			SetCritical(0);
		}

	|	expr TOKEN_XOR expr
		{
			SetCritical(-1);
			if ( (anInt = IsSimple($1)) == -1) {
				error(ERR_NONFATAL, EXPERR);
				YYERROR;
			};

			if ( anInt != IsSimple($3) ) {
				error(ERR_NONFATAL, TYPERR);
				YYERROR;
			};

			if ( anInt != INTEGER ) {
				error(ERR_NONFATAL, OPRERR);
				YYERROR;
			};

			anInt= IntVal($1) ^ IntVal($3);
		    $$ = Vectorize(INTEGER, 1,  anInt);
			SetCritical(0);
		}
	;

expr0	:	TOKEN_ID
		{
			if ( IDType($1) == LABEL ) {
			    if ( IDFlag($1) == LOCAL ) {
				    /* Reference to a local procedure variable */
				    $$ = Vectorize((CodeBits == 16 ? R_BP : R_EBP), 0);
					Vectorize(INTEGER, 1, $1->Value);
				} else if ( IDFlag($1) == GLBLBL &&
							$1->segment == UNDEF_SEG ) {
					/* Global symbol is still undefined, return
					   a forward reference */
					$$ = Vectorize(SYMREF, 1, $1);
				} else {
					/* Reference to a defined symbol. */
					$$= Vectorize(SEGMENT, 0, $1->segment);
					Vectorize(INTEGER, 1, $1->offset);
				};
			} else if ( $1->Type == NO_TYPE ) {
				/* Forward reference */
				if ( IsCritical() ) {
					/* If flagCritical is set forward references
					   are not allowed */
					error(ERR_NONFATAL, UNDERR, $1->Name);
					FreeSymbol($1);
					YYERROR;
				};

				$$ = Vectorize(SYMREF, 1, $1);
			} else {
			/* We should never reach this, if so panic. */
				error(ERR_PANIC, GENERR, "parser error 1.");
			};
		}

	|	TOKEN_REG   { $$ = Vectorize($1, 1); }
	|	TOKEN_NUM   { $$ = Vectorize(INTEGER, 1, $1); }
	|	TOKEN_FLT   { $$ = Vectorize(FLOAT, 1, $1); }
	|	TOKEN_STR   { $$ = Vectorize(STR, 1, $1, TokenInfo); }
	|	TOKEN_CHR   { $$ = Vectorize(CHAR, 1, $1); }
	|	TOKEN_HERE  { $$ = Vectorize(INTEGER, 1, Offset + LocalOffset); }
	|	TOKEN_BASE  { $$ = Vectorize(INTEGER, 1, Offset); }
	;

/* -------- Boolean expressions parsing ------------------------------- */

bexpr : bexpr0
		{ $$= $1; }
	|	'(' bexpr ')' TOKEN_AND '(' bexpr ')'
		{ $$ = ($2 && $6); }
	|	'(' bexpr ')' TOKEN_OR '(' bexpr ')'
		{ $$ = ($2 || $6 ); }
	|	TOKEN_NOT '(' bexpr ')'
		{ $$ = !$3; }
	;

bexpr0 : TOKEN_DEF '(' TOKEN_ID ')'
		{
			if ( $3->Type == NO_TYPE ) {
				$$= FALSE;
				FreeSymbol($3);
			} else {
				$$= TRUE;
			}
		}
	|	expr TOKEN_EQ expr
		{
			SetCritical(-1);

			if ( (anInt = IsSimple($1)) == -1 ) {
				error(ERR_NONFATAL, EXPERR);
				YYERROR;
			} else if ( anInt != IsSimple($3) ) {
				error(ERR_NONFATAL, TYPERR);
				YYERROR;
			};

			switch ( anInt ) {
				case INTEGER:
				case CHAR:
					anInt= (IntVal($1) == IntVal($3));
					break;

				case FLOAT:
					anInt= (FloatVal($1) == FloatVal($3));
					break;

				case STR:
					anInt= strcmp(StrVal($1), StrVal($3));
			};

			/* if $2 is zero we are checking for equality */
			$$= ( $2 ? anInt : !anInt);
			SetCritical(0);
		}

	|	expr TOKEN_CMP expr
		{
			SetCritical(-1);

			if ( (anInt = IsSimple($1)) == -1 ) {
				error(ERR_NONFATAL, EXPERR);
				YYERROR;
			} else if ( anInt != IsSimple($3) ) {
				error(ERR_NONFATAL, TYPERR);
				YYERROR;
			};

			/* First check for equality. */
			switch ( anInt ) {
				case INTEGER:
				case CHAR:
					$$= (IntVal($1) == IntVal($3));
					break;

				case FLOAT:
					$$= (FloatVal($1) == FloatVal($3));
					break;

				case STR:
					$$= (strcmp(StrVal($1), StrVal($3)) == 0);
			};

			if ( !$$ || ( $2 != 2 && $2 != 4 ) ) {
			/* If not equal or strong comparation. */
				switch ( anInt ) {
					case INTEGER:
					case CHAR:
						anInt= (IntVal($1) < IntVal($3));
						break;

					case FLOAT:
						anInt= (FloatVal($1) < FloatVal($3));
						break;

					case STR:
						anInt= (strcmp(StrVal($1), StrVal($3)) < 0);
				};

				$$= ( $2 == 1 || $2 == 3 ? anInt : !anInt );
			};

			SetCritical(0);
		}
	;

%%

/* -------- Local Functions ------------------------------------------- */

/*
	Function......: AssembleProcProlog
	Description...: Generates code for a subroutine prolog.
					Returns the size of the generated code.
	Parameters....:
					Proc		Pointer to the procedure info struct.

					The struct defines the following fields:

					Id			pointer the the symbol of the
								procedure
					SavedRegs	Array containing the codes of registers
								used by the procedure, they must be saved
								on the stack.
					RegCount	Number of rregister to be saved.
					LocalSize	Size of local variables.
					ParamSize	Size of local parameters. */
static long AssembleProcProlog(struct ProcInfo *Proc)
{
	INSN	dummy;
	register int i;
	register long retval= 0;

	/* Push BP/EBP register if there are local variable or parameter
	   and intialize the BP/EBP register with the value in SP/ESP */
	if ( Proc->LocalSize || Proc->ParamSize ) {
	    bzero(&dummy, sizeof(INSN));
	    dummy.times	  = 1;
	    dummy.opcode  = I_PUSH;
	    dummy.operands= 1;

		i = ( CodeBits == 16 ? R_BP : R_EBP);
		dummy.oprs[0].type	 = REGISTER;
	    dummy.oprs[0].type	 |= reg_flags[i];
		dummy.oprs[0].basereg = i;

		retval += assemble(Segment, Offset + retval, CodeBits, &dummy);

		bzero(&dummy, sizeof(INSN));
		dummy.times	  = 1;
		dummy.opcode  = I_MOV;
		dummy.operands= 2;

		dummy.oprs[0].type	 = REGISTER;
	    dummy.oprs[0].type	 |= reg_flags[i];
		dummy.oprs[0].basereg = i;

		i= ( CodeBits == 16 ? R_SP : R_ESP);
		dummy.oprs[1].type	 = REGISTER;
	    dummy.oprs[1].type	 |= reg_flags[i];
		dummy.oprs[1].basereg = i;

		retval += assemble(Segment, Offset + retval, CodeBits, &dummy);
	};

	/* Reserve space for locals if any */
	if ( Proc->LocalSize ) {
		bzero(&dummy, sizeof(INSN));
		dummy.times	  = 1;
		dummy.opcode  = I_SUB;
		dummy.operands= 2;

		i= ( CodeBits == 16 ? R_SP : R_ESP);
		dummy.oprs[0].type	 = REGISTER;
	    dummy.oprs[0].type	 |= reg_flags[i];
		dummy.oprs[0].basereg = i;

		dummy.oprs[1].type	  = IMMEDIATE;
		dummy.oprs[1].segment = NO_SEG;
		dummy.oprs[1].wrt	  = NO_SEG;
		dummy.oprs[1].offset  = Proc->LocalSize;

		retval += assemble(Segment, Offset + retval, CodeBits, &dummy);
	};

	/* Push into the stack the registers to be saved. */
	for (i= 0; i < Proc->RegCount; i++) {
		bzero(&dummy, sizeof(INSN));
	    dummy.times	  = 1;

		if ( Proc->RegList[i] < REG_ENUM_LIMIT ) {
			dummy.opcode  = I_PUSH;
	        dummy.operands= 1;
		    dummy.oprs[0].type	 = REGISTER;
	        dummy.oprs[0].type	 |= reg_flags[Proc->RegList[i]];
		    dummy.oprs[0].basereg = Proc->RegList[i];
		} else {
			dummy.opcode  = ( Proc->RegList[i] == REG_ENUM_LIMIT ?
				I_PUSHAD : I_PUSHAW );
			dummy.operands= 0;
		}

		retval += assemble(Segment, Offset + retval, CodeBits, &dummy);
	};

	return retval;
}


/*
	Function......: AssembleProcEpilog
	Description...: Generates code for a subroutine epilog.
					Returns the size of the generated code.
	Parameters....:
					Proc			Pointer to the procedure info
									structure. */
static long AssembleProcEpilog(struct ProcInfo *Proc)
{
	INSN	dummy;
	register int i;
	register long retval= 0;

	/* No defined procedure, we have no code to assemble */
	if ( !Proc || !Proc->Id ) return 0;

	/* Pop out of the stack the saved registers. */

	for (i= Proc->RegCount - 1; i >= 0; i--) {
		bzero(&dummy, sizeof(INSN));
	    dummy.times	  = 1;

		if ( Proc->RegList[i] < REG_ENUM_LIMIT ) {
			dummy.opcode  = I_POP;
	        dummy.operands= 1;
		    dummy.oprs[0].type	 = REGISTER;
	        dummy.oprs[0].type	 |= reg_flags[Proc->RegList[i]];
		    dummy.oprs[0].basereg = Proc->RegList[i];
		} else {
			dummy.opcode  = ( Proc->RegList[i] == REG_ENUM_LIMIT ?
				I_POPAD : I_POPAW );
			dummy.operands= 0;
		}

		retval += assemble(Segment, Offset + retval, CodeBits, &dummy);
	};

	/* Pop the BP register or assemble the 'leave' instruction. */
	if ( Proc->LocalSize || Proc->ParamSize ) {
	    bzero(&dummy, sizeof(INSN));
		if ( CodeBits == 16 ) {

			if ( Proc->LocalSize ) {
				/* mov sp, bp */
		        dummy.times	  = 1;
		        dummy.opcode  = I_MOV;
		        dummy.operands= 2;

			    dummy.oprs[0].type	 = REGISTER;
		        dummy.oprs[0].type	 |= reg_flags[R_SP];
			    dummy.oprs[0].basereg = R_SP;

			    dummy.oprs[1].type	 = REGISTER;
		        dummy.oprs[1].type	 |= reg_flags[R_BP];
			    dummy.oprs[1].basereg = R_BP;

			    retval += assemble(Segment, Offset + retval, CodeBits,
					&dummy);
			};

			/* pop bp */
	        bzero(&dummy, sizeof(INSN));
	        dummy.times	  = 1;
	        dummy.opcode  = I_POP;
	        dummy.operands= 1;

		    dummy.oprs[0].type	 = REGISTER;
	        dummy.oprs[0].type	 |= reg_flags[R_BP];
		    dummy.oprs[0].basereg = R_BP;
		} else {
	        dummy.times	  = 1;
	        dummy.opcode  = I_LEAVE;
		}

		retval += assemble(Segment, Offset + retval, CodeBits, &dummy);
	};

	return retval;
}


/*
	Function......: IncludeBinary
	Description...: Include a binary file.
					Retunrs the number of bytes successfully included
					into the assembled code.
	Parameters....:
					Filename		Name of the binary file to include
					TotalSize		Total bytes to be included
					Skip			Byte to skip in the source binary
									file. */
static long IncludeBinary(char *Filename, long TotalSize, long Skip)
{
	register long Count= 0;
	register long BytesRead, Bytes2Read;
	FILE *Source;
	char Buffer[1024];

	/* Return if there are no bytes to insert */
	if ( !TotalSize ) return 0;

	/* Open source file */
	LIB_setflags(ERR_NONFATAL);
	Source= LIB_fopen(Filename, M_RDBIN);
	LIB_setflags( -1 ); // Restores the default error reporting flags

	if ( !Source ) return 0;

	/* Assing a value to TotalSize */
	if ( TotalSize == -1 ) {
		fseek(Source, 0L, SEEK_END);
		TotalSize= ftell(Source);
	};

	/* Skip the requested number of bytes */
	fseek(Source, Skip, SEEK_SET);

	do {
		/* Calculates the number of bytes to read in. */
		Bytes2Read= TotalSize - Count;
		if ( Bytes2Read > sizeof(Buffer) ) Bytes2Read= sizeof(Buffer);

		/* Read a chunk of bytes from source file ... */
		BytesRead= fread(Buffer, 1, Bytes2Read, Source);
		/* ... and lay them down on the assembled segment */
		Output(Segment, Offset + Count, Buffer, OUT_RAWDATA + BytesRead,
			NO_SEG, NO_SEG);

		Count += BytesRead;
	} while ( BytesRead == Bytes2Read && Count < TotalSize );

	fclose(Source);

	/* Pad if we haven't fullfilled the TotalSize request. */
	if ( Count < TotalSize ) {
		pad(Segment, Offset + Count, 0, TotalSize - Count);
	};

	return TotalSize;
}


/*
	Function......: DefineLabel
	Description...: Performs all the operation necessary to a label
					definition:
						- Initializes the fields of the struct.
						- Defines the symbol in the output format.
						- Performs the backpatch of the symbol references.
					Returns 0 if successful, -1 otherwise.
					On error calls the global error report function.
	Parameters....:
					Sym     pointer to the SYMBOL struct to be defined. */
static int DefineLabel(SYMBOL *Sym)
{
	int isLocalLabel = !LIB_strnicmp("@@", Sym->Name, 2);

	/* Fill in the SYMBOL fields */
	Sym->Value	= Offset;
	Sym->segment= Segment;
	Sym->offset	= Offset;

	/* Declaration of a local label */
	if (  isLocalLabel ) {
		if ( !CurrProc.Id ) {
		/* Link the symbol to the current section */
		    LinkSymbol(SecId, Sym);
		} else {
		/* Link the local symbol to the closest
		   procedure declaration */
			LinkSymbol(CurrProc.Id, Sym);
		};

	};

	/* Definition of the symbol in the output format */
	if ( Sym->Type == NO_TYPE ) {
	/* Definition of a symbol local to the module */
		Sym->Type = LABEL;
		Symdef(Sym, 0);
	} else if ( IDFlag(Sym) == GLBLBL && !isLocalLabel ) {
	/* Definition of a global symbol */
		Symdef(Sym, 1);
	} else {
		error(ERR_NONFATAL, LBLERR, Sym->Name);
		return -1;
	};

	/* Backpatch the symbol! */
	BPatch(Sym);
	return 0;
}


/*
	Function......: CalcMemRef
	Description...: Calculates the memory reference specified by its
					segment and offset part, converting it to a valid
					operand.
					Returns a pointer to a static storage area, overwritten
					on each call, containing the operand if successfull,
					NULL otherwise.
	Parameters....:
					Seg			pointer to the segment part expression.
					Offs		pointer to the offset part expression. */
static OPERAND *CalcMemRef(register EXPR_RESULT *Seg,
	register EXPR_RESULT *Offs)
{
	static OPERAND anOp;
	int b, i, s;	   /* basereg, indexreg, scale */
	long o;			   /* offset */

	/* Collects forward references if any */
	SymRefs(Offs, OprsCount);

	bzero(&anOp, sizeof(OPERAND));

	if ( Seg ) {	/* segment override */
		if ( Seg[1].type != EXPR_END || Seg->int_result != 1 ||
			 REG_SREG & ~reg_flags[Seg->type] ) {
			error(ERR_NONFATAL, SEGERR);
			return NULL;
		} else if (instr.nprefix == 3) {
			error(ERR_NONFATAL, PREFSERR);
			return NULL;
		} else {
			instr.prefixes[instr.nprefix++] = Seg->type;
		};
	}

	b = i = -1;
	o = s = 0;

	if ( Offs->type < REG_ENUM_LIMIT ) {	/* this bit's a register */
		if ( Offs->int_result == 1 ) {
		/* in fact it can be basereg */
			b = Offs->type;
		} else {
		/* no, it has to be indexreg */
			i = Offs->type;
			s = Offs->int_result;
		};
		Offs++;
	}

	if ( Offs->type != EXPR_END && Offs->type < REG_ENUM_LIMIT ) {
	/* it's a second register */

		if ( Offs->int_result != 1 ) {
		/* it has to be indexreg */
			if ( i != -1 ) {	 /* but it can't be */
				error(ERR_NONFATAL, EAERR);
				return NULL;
			} else {
				i = Offs->type;
				s = Offs->int_result;
			};
		} else {
		/* it can be basereg */
			if ( b != -1 ) {
			/* or can it? */
				i = Offs->type;
				s = 1;
			} else {
				b = Offs->type;
			};
		}
		Offs++;
	};

	if ( Offs->type != EXPR_END ) { /* is there an offset? */
		if ( IsSimple(Offs) != INTEGER ) { /* in fact, is there an error? */
			error(ERR_NONFATAL, EAERR);
			return NULL;
		} else {
			anOp.offset_flag= TRUE;
			o = IntVal(Offs);
		}
	}

	anOp.type = MEMORY;

	if ( b == -1 && ( i == -1 || s == 0) ) {
		anOp.offset_flag= TRUE;
		anOp.type |= MEM_OFFS;
	};

	anOp.basereg  = b;
	anOp.indexreg = i;
	anOp.scale	  = s;
	anOp.wrt	  = NO_SEG;
	anOp.offset	  = o;
	anOp.segment  = SegmentVal(Offs);

	EvaluatorReset();
	mref= 0;

	return &anOp;
}


/*
	Function......: ProcessOverride
	Description...: Processes operand size override specifier.
					Returns the appropriate operand type flag if
					successfull, -1 otherwise.
	Parameters....:
					SizeSpec		  Value of the specifier. */
static int ProcessOverride(int SizeSpec)
{
	switch ( SizeSpec ) {
		case S_BYTE:	return BITS8;
		case S_WORD:	return BITS16;
		case S_DWORD:	return BITS32;
		case S_QWORD:	return BITS64;
		case S_TWORD:	return BITS80;
		case S_FAR:		return FAR;
		case S_NEAR:	return NEAR;
		case S_SHORT:	return SHORT;
	};

	error(ERR_NONFATAL, INVSIZERR);
	return -1;
}


/*
	Function......: ProcessOperand
	Description...: Processes an expression converting it in a valid operand
					for an istructions.
					Returns a pointer to static a storage area, overwritten
					on each call, containing the operand if successfull,
					NULL if errors are encountered.
	Parameters....:
					e			pointer to the expression containing the
								operand. */
static OPERAND *ProcessOperand(EXPR_RESULT *e)
{
	static OPERAND anOp;

	if ( mref ) {
	/* If the global flag 'mref' is set the expression is a memory reference
	   so call the appropriate function to resolve it */
		return CalcMemRef(NULL, e);
	}

	bzero(&anOp, sizeof(OPERAND));
	anOp.segment  = NO_SEG;
	anOp.wrt	  = NO_SEG;
	anOp.basereg  = -1;

	switch ( IsSimple(e) ) {
		case INTEGER:
		/* Immediate operand */
			anOp.type	 |= IMMEDIATE;
			anOp.offset	 = IntVal(e);
			anOp.segment = SegmentVal(e);

			if ( !SymRefs(e, OprsCount) &&
				 anOp.offset == 1 && anOp.segment == NO_SEG ) {
			/* The operand is the numeric constant 1 */
				anOp.type |= UNITY;
			}
			break;

		case CHAR:
		/* Immediate char constant with any symbol reference */
			anOp.type	 |= IMMEDIATE;
			anOp.offset	 = ChrVal(e);
			break;

		case FLOAT:
		case STR:
			error(ERR_NONFATAL, INSERR);
			return NULL;

		default:
			if ( e->type >= MAXTYPE || e->int_result != 1 ) {
				error(ERR_NONFATAL, INSERR);
				return NULL;
			};

			/* Immediate register */
			anOp.type	&= TO;
			anOp.type	|= REGISTER;
			anOp.type	|= reg_flags[e->type];
			anOp.basereg = e->type;

			/* There would not be nothing after the register. */
			e++;
			if ( e->type != EXPR_END ) {
				error(ERR_NONFATAL, INSERR);
				return NULL;
			};
	};

	EvaluatorReset();
	return &anOp;
}


/*
	Function......: ProcessExtop
	Description...: Builds up an operand for a data definition istruction
					starting from the expression that defines the operand.
					Returns the pointer to the constructed operand if
					successfull, NULL otherwise. The function reports errors
					via the global error reporting function.
					NOTE: This function modifies the global variable
						  LocalOffset.
	Parameters....:
					Code		opcode of the data definition istruction.
					e			pointer to the expression to start from. */
static EXTOP *ProcessExtop(int Code, EXPR_RESULT *e)
{
	EXTOP *RetVal;
	int type;
	short int DataSize;
	short int Alignment;

	type= IsSimple(e);

	if ( type == -1 || Code < I_DB || Code > I_DW ) {
		error(ERR_NONFATAL, INSERR);
		EvaluatorReset();
		return NULL;
	};

	/* Build up the return value.
	   Datas are stored in the EXTOP structure only in two format:
		   - EOT_DB_NUMBER	a single number (integer or char)
			 stored in the offset field of the structure.
		   - EOT_DB_STRING	a string of datas (chars or float values)
			 stored in a memory area pointed by the stringval field of
			 the structure whose size is specified by the stringlen field
			 of the same structure.
	   The code of the data definition pseudo instruction tells only which
	   alignment to use for data layout. */
	RetVal= (EXTOP *)LIB_malloc(sizeof(EXTOP));

	switch ( type ) {
		case INTEGER:
			RetVal->type     = EOT_DB_NUMBER;
			RetVal->extop_off= IntVal(e);

			if ( Code == I_DB && RetVal->extop_off >= 0 && RetVal->extop_off <= 255 ) {
				DataSize= 1;
			    RetVal->extop_seg= NO_SEG;
	            RetVal->extop_wrt= NO_SEG;
			} else if ( RetVal->extop_off >= -32768 && RetVal->extop_off <= 32767 ) {
				DataSize= sizeof(short int);
	            RetVal->extop_seg= SegmentVal(e);
	            RetVal->extop_wrt= NO_SEG;
			} else {
				DataSize= sizeof(long);
	            RetVal->extop_seg= SegmentVal(e);
	            RetVal->extop_wrt= NO_SEG;
			};
			break;

		case CHAR:
			RetVal->type     = EOT_DB_NUMBER;
			RetVal->extop_off= ChrVal(e);
	   	    RetVal->extop_seg= NO_SEG;
			RetVal->extop_wrt= NO_SEG;
			DataSize= 1;
			break;

		case FLOAT:
			RetVal->type	    = EOT_DB_STRING;
			RetVal->extop_string= (char *)LIB_malloc(sizeof(float));
			DataSize= sizeof(float);
			/* Yes, this is the power of C ! */
			*((float *)RetVal->extop_string)= FloatVal(e);
			break;

		case STR:
			RetVal->type	     = EOT_DB_STRING;
			RetVal->extop_string = StrVal(e);
			RetVal->extop_length = e->str_leng;
			DataSize= e->str_leng;
	};

	/* Now collect forward references contained in the expression */
	if ( RetVal->type == EOT_DB_NUMBER &&
		 SymRefs(e, -1, RetVal->extop_off, RetVal->extop_seg) ) {
		/* The expression contains forward reference so it can be assembled
		   the returned operand is declared of type DUMMY, it is used only
		   to calculate istruction size and it not actually generate
		   code. */
		RetVal->type = EOT_DUMMY;
	};

	/* Now align datas and modify LocalOffset */
	switch ( Code ) {
		case I_DB:
			Alignment= 1;
			break;

		case I_DW:
			Alignment= 2;
			break;

		case I_DD:
			Alignment= 4;
			break;

		case I_DQ:
			Alignment= 8;
			break;

		case I_DT:
			Alignment= 10;
			break;

		default:
			error(ERR_FATAL, GENERR, "parser error in ProcessExtop().");
			return NULL;
	};

	LocalOffset += - (DataSize % Alignment);

	EvaluatorReset();
	return RetVal;
}


/*
	Function......: InstrCleanup
	Description...: Clears all the fields of the instr struct, resets some
					global variable, preparing the parser for the next
					instruction.
	Parameters....: None. */
static void InstrCleanup()
{
	register EXTOP *Crsr = instr.eops;
	register EXTOP *Next;

	/* Clear extops if any */
	while ( Crsr ) {
		Next = Crsr->next;
		LIB_free(Crsr);
		Crsr = Next;
	};

	bzero(&instr, sizeof(INSN));
	instr.times		=  1;
	instr.opcode	= -1;
	instr.condition = -1;

	LocalOffset = 0;
	mref		= 0;
	OprsCount	= 0;
	PatchPtr	= NULL;
	EvaluatorReset();
}


/*
	Function......: SymRefs
	Description...: Collects all forward references within an expression.
					Returns the number of forward references encountered.
					The function also adds the patch information to the
					list of references associated with a symbol.
	Parameters....:
					e				pointer to the expression to be analyzed
					PatchType		type of the patch informations to add
					...				additional informations. */
static int SymRefs(register EXPR_RESULT *e, short PatchType, ...)
{
	register int RetVal = 0;
	va_list	 arg_ptr;

	va_start(arg_ptr, PatchType);

	while ( e->type != EXPR_END ) {

		if ( e->type == SYMREF ) {
			RetVal++;
			if ( PatchPtr == NULL ) {
			/* Allocate the patch which the reference will referes to */
				PatchPtr= (PATCH *)LIB_malloc(sizeof(PATCH));
				PatchPtr->Seg	 = Segment;

				if ( PatchType < 0 || PatchType >= 3 ) {
				    PatchPtr->Offset    = LocalOffset + Offset;
					PatchPtr->patch_off = va_arg(arg_ptr, long);
					PatchPtr->patch_seg = va_arg(arg_ptr, long);
				} else {
					PatchPtr->Offset = Offset;
				};

				PatchPtr->Bits	 = CodeBits;
			};

			/* Add the reference to the backpatchlist of the symbol */
			AddPatchRef(e->SymId, PatchPtr, PatchType, e->SymOccur);
		};

		e++;
	};

	va_end(arg_ptr);
	return RetVal;
}

static int yyerror(char *s)
{
	error(ERR_NONFATAL, PARSE_ERR, yytext);
	return -1;
}

/* -------- Public Functions ------------------------------------------ */




/*
	Function......: ParserInit
	Description...: Initializes the parser setting global variables and
					opening the input file.
					Parses the input file and generates the output calling
					the _cleanup() dirver interface function.
					This function is the main entry point of the whole
					module.
					Returns 0 if the parser has been correctly initialized
					(i.e. the file was correctly opened), -1 on errors.
					The functions calls the global error reporting function
					if errors are detected.
					NOTE: Each parameter is a pointer to an area of static
						  storage of size FILENAME_MAX.
	Parameters....:
					InName			Name of the file to be parsed.
					OutName			Name of the file to be generated. */
int Compile(char *InName, char *OutName)
{
	int undefined;

	instr.eops = NULL;
	InstrCleanup();
	yylineno	= 1;
	CodeBits	= 16;

	bzero(&CurrProc, sizeof(struct ProcInfo));
	LIB_addExtension(InName, InName, ".ASM");

	/* Setup the input buffer */
	if ( SwitchToFile(InName) ) return -1;

	/* Initialize the output driver */
	if ( DriverInit(OutName) ) return -1;

	/* Define the symbol for the text section (required) and
	   switch to this section. */
	SecId = LookUp("TEXT", IdCase);
	SecId->Type		= (LABEL | SECLBL);
	SecId->Value	= -1;
	SecId->segment	= Segment = Section(NULL);
	SecId->offset	= Offset  = 0;

	/* Call the parsing function */
	if ( yyparse() || Errors ) {
		error(ERR_FATAL | ERR_NOFILE, EXITERR, Errors);
		return -1;
	};

	if ( (undefined= UndefSymNum()) ) {
		error(ERR_FATAL | ERR_NOFILE, UNDEFSYMSERR, undefined);
		return -1;
	};

	Cleanup();

	return 0;
}

/* --------- End of File --------------------------------------- */
