/*
		JAS v1.00							  (c) 1996 NikSoft International
		--------------------------------------------------------------------
		PTYPES.H

		Types and costant definition used by the lexical analyzer and the
		parser.

		04/11/96 18.52	Version v1.00
			Basically the types managed by the parser are the same as
			the one treated by the NASM parser.
*/

/* ------------------------------------------------------------- */

#ifndef __PTYPES_H
#define __PTYPES_H

/* operand to an instruction */
typedef struct operand {
	long    type;						/* type of operand */
	short	offset_flag;				/* flags if the offset field is
										   valid. */
	short   addr_size;					/* 0 means default; 16; 32 */
	int     basereg, indexreg, scale;	/* registers and scale involved */
	long    segment;					/* immediate segment, if needed */
	long    offset;						/* any immediate number */
	long    wrt;						/* segment base it's relative to */

} OPERAND;


/* extended operand */
typedef struct extop {
	struct extop *next;			/* linked list */
	long    type;				/* defined below */

	union {
	    struct {
			char	*__stringval;	/* if it's a string, then here it is */
	        short   __stringlen;    /* ... and here's how long it is */
		}	__string_op;

   		struct {
	        long    __segment;		/* if it's a number/address, then... */
	        long    __offset;		/* ... it's given here ... */
	        long    __wrt;			/* ... and here */
		}	__number_op;
	}		__extop_datas;
} EXTOP;

/* Macro definition to ease handling of the previuos struct */
#define extop_seg			__extop_datas.__number_op.__segment
#define extop_off			__extop_datas.__number_op.__offset
#define extop_wrt			__extop_datas.__number_op.__wrt

#define extop_string		__extop_datas.__string_op.__stringval
#define extop_length		__extop_datas.__string_op.__stringlen


/* an instruction itself */
typedef struct insn {
	struct symbol *label;	/* the label defined, or NULL */
	short   prefixes[2];	/* instruction prefixes, if any */
	short   nprefix;		/* number of entries in above */
	short   opcode;			/* the opcode - not just the string */
	short   condition;		/* the condition code, if Jcc/SETcc */
	short   operands;		/* how many operands? 0-3 */
	OPERAND oprs[3];		/* the operands, defined as above */
	EXTOP	*eops;			/* extended operands */
	short   times;			/* repeat count (TIMES prefix) */

	char	*template_code; /* Template code that match the
							   instruction, it is used at
						       assemble time */
	long	code_size;      /* Size of the code generated for
							   the instruction. */
} INSN;


/* the result of an expression */
typedef struct expr_result {
	long            type;
	union {
		long	_int_result;	 /* What it is */
		double	_flt_result;
		struct	{
			char *_stringdata;
			int	 _stringlen;
		}		_string;
		char	_chr_result;
		struct {				/* To keep track of forward references */
			struct symbol *_Id; /* Symbol table address of the referenced
								   symbol */
			int			  _occur;	/* Occurences of the symbol within the
									   expression */
		}       _symref;
	}				_value;
} EXPR_RESULT;


/* Some useful macro's definition to handle the fields of the struct
   defined above */
#define int_result				_value._int_result
#define flt_result				_value._flt_result
#define str_result				_value._string._stringdata
#define str_leng				_value._string._stringlen
#define chr_result				_value._chr_result
#define SymId					_value._symref._Id
#define SymOccur				_value._symref._occur


/* Next, the codes returned from the parser, for registers and
   instructions. */
enum {				       /* register names */
	R_AH = 1, R_AL, R_AX, R_BH, R_BL, R_BP, R_BX, R_CH, R_CL, R_CR0,
	R_CR2, R_CR3, R_CR4, R_CS, R_CX, R_DH, R_DI, R_DL, R_DR0, R_DR1,
	R_DR2, R_DR3, R_DR6, R_DR7, R_DS, R_DX, R_EAX, R_EBP, R_EBX,
	R_ECX, R_EDI, R_EDX, R_ES, R_ESI, R_ESP, R_FS, R_GS, R_SI, R_SP,
	R_SS, R_ST0, R_ST1, R_ST2, R_ST3, R_ST4, R_ST5, R_ST6, R_ST7,
	R_TR3, R_TR4, R_TR5, R_TR6, R_TR7, REG_ENUM_LIMIT
};

/* Expression type costants */
enum {
	SYMREF = REG_ENUM_LIMIT,
	INTEGER, FLOAT, STR, CHAR,	/* Expressions of these types are called
								   simple expression */
	SEGMENT,
	MAXTYPE, EXPR_END = -1
};

enum {				       /* instruction names */
	I_AAA, I_AAD, I_AAM, I_AAS, I_ADC, I_ADD, I_AND, I_ARPL,
	I_BOUND, I_BSF, I_BSR, I_BSWAP, I_BT, I_BTC, I_BTR, I_BTS,
	I_CALL, I_CBW, I_CDQ, I_CLC, I_CLD, I_CLI, I_CLTS, I_CMC, I_CMP,
	I_CMPSB, I_CMPSD, I_CMPSW, I_CMPXCHG, I_CMPXCHG8B, I_CPUID,
	I_CWD, I_CWDE, I_DAA, I_DAS, I_DEC, I_DIV, I_ENTER, I_F2XM1,
	I_FABS, I_FADD, I_FADDP, I_FBLD, I_FBSTP, I_FCHS, I_FCLEX, I_FCMOVB,
	I_FCMOVBE, I_FCMOVE, I_FCMOVNB, I_FCMOVNBE, I_FCMOVNE, I_FCMOVNU,
	I_FCMOVU, I_FCOM, I_FCOMI, I_FCOMIP, I_FCOMP, I_FCOMPP, I_FCOS,
	I_FDECSTP, I_FDISI, I_FDIV, I_FDIVP, I_FDIVR, I_FDIVRP, I_FENI,
	I_FFREE, I_FIADD, I_FICOM, I_FICOMP, I_FIDIV, I_FIDIVR, I_FILD,
	I_FIMUL, I_FINCSTP, I_FINIT, I_FIST, I_FISTP, I_FISUB, I_FISUBR,
	I_FLD, I_FLD1, I_FLDCW, I_FLDENV, I_FLDL2E, I_FLDL2T, I_FLDLG2,
	I_FLDLN2, I_FLDPI, I_FLDZ, I_FMUL, I_FMULP, I_FNOP, I_FPATAN,
	I_FPREM, I_FPREM1, I_FPTAN, I_FRNDINT, I_FRSTOR, I_FSAVE,
	I_FSCALE, I_FSETPM, I_FSIN, I_FSINCOS, I_FSQRT, I_FST, I_FSTCW,
	I_FSTENV, I_FSTP, I_FSTSW, I_FSUB, I_FSUBP, I_FSUBR, I_FSUBRP,
	I_FTST, I_FUCOM, I_FUCOMI, I_FUCOMIP, I_FUCOMP, I_FUCOMPP,
	I_FXAM, I_FXCH, I_FXTRACT, I_FYL2X, I_FYL2XP1, I_HLT, I_ICEBP,
	I_IDIV, I_IMUL, I_IN, I_INC, I_INSB, I_INSD, I_INSW, I_INT,
	I_INT1, I_INT01, I_INT3, I_INTO, I_INVD, I_INVLPG, I_IRET,
	I_IRETD, I_IRETW, I_JCXZ, I_JECXZ, I_JMP, I_LAHF, I_LAR, I_LDS,
	I_LEA, I_LEAVE, I_LES, I_LFS, I_LGDT, I_LGS, I_LIDT, I_LLDT,
	I_LMSW, I_LOADALL, I_LODSB, I_LODSD, I_LODSW, I_LOOP, I_LOOPE,
	I_LOOPNE, I_LOOPNZ, I_LOOPZ, I_LSL, I_LSS, I_LTR, I_MOV,
	I_MOVSB, I_MOVSD, I_MOVSW, I_MOVSX, I_MOVZX, I_MUL, I_NEG,
	I_NOP, I_NOT, I_OR, I_OUT, I_OUTSB, I_OUTSD, I_OUTSW, I_POP,
	I_POPA, I_POPAD, I_POPAW, I_POPF, I_POPFD, I_POPFW, I_PUSH,
	I_PUSHA, I_PUSHAD, I_PUSHAW, I_PUSHF, I_PUSHFD, I_PUSHFW, I_RCL,
	I_RCR, I_RDMSR, I_RDPMC, I_RDTSC,  I_RET, I_RETF, I_RETN, I_ROL,
	I_ROR, I_RSM, I_SAHF, I_SAL, I_SALC, I_SAR, I_SBB, I_SCASB, I_SCASD,
	I_SCASW, I_SGDT, I_SHL, I_SHLD, I_SHR, I_SHRD, I_SIDT, I_SLDT, I_SMSW,
	I_STC, I_STD, I_STI, I_STOSB, I_STOSD, I_STOSW, I_STR, I_SUB,
	I_TEST, I_UMOV, I_VERR, I_VERW, I_WAIT, I_WBINVD, I_WRMSR,
	I_XADD, I_XCHG, I_XLATB, I_XOR, I_CMOVcc, I_Jcc, I_SETcc,

	/* Pseudo istructions codes... don't modify their position and order! */
	I_RESB, I_RESD, I_RESQ, I_REST,	I_RESW,
	I_DB, I_DD, I_DQ, I_DT, I_DW,

	/* Non valid or not implemented istruction code are defined
	   beyond this line */
	MAX_CODE, I_EQU
};

enum {				       /* condition code names */
	C_A, C_AE, C_B, C_BE, C_C, C_E, C_G, C_GE, C_L, C_LE, C_NA, C_NAE,
	C_NB, C_NBE, C_NC, C_NE, C_NG, C_NGE, C_NL, C_NLE, C_NO, C_NP,
	C_NS, C_NZ, C_O, C_P, C_PE, C_PO, C_S, C_Z
};

enum {						/* special tokens */
	S_BYTE, S_DWORD, S_FAR, S_LONG, S_NEAR, S_QWORD, S_SHORT, S_TO,
	S_TWORD, S_WORD
};

/* Note that because segment registers may be used as instruction
   prefixes, we must ensure the enumerations for prefixes and
   register names do not overlap. */
enum {				       /* instruction prefixes */
	PREFIX_ENUM_START = REG_ENUM_LIMIT,
	P_LOCK = PREFIX_ENUM_START, P_REP, P_REPE, P_REPNE, P_REPNZ, P_REPZ,
};

enum {				       /* extended operand types */
	EOT_DUMMY, EOT_DB_STRING, EOT_DB_NUMBER
};

/* Parser initialization function */
extern int Compile(char *, char *);
#endif

/* --------- End of File --------------------------------------- */
