/*
		JAS v1.00							  (c) 1997 NikSoft International
		--------------------------------------------------------------------
		SYMMNGR.H

		Symbol table management module header file.
		Type and function prototypes definitions.

		06/11/96 19.58	Version 1.00
*/

#ifndef __SYMMNGR_H
#define __SYMMNGR_H

/* Symbol type costant.
   They use the high word of the Type field of the struct, the low one is
   used by the general type system costants defined in the file
   'globals.h'.
   The lower nibble of the high word is used to store the type class of
   the symbol, i.e. a label, a macro definition, a place holder ...
   while the higher stores various type flags concerning the type class
   of the symbol, for example for a label it tells if it is a label for
   a local parameter, for a section name and so on. */

/* Type class const */
#define NO_TYPE				0xFFFF0000L
#define LABEL				0x00010000L
#define MACRO				0x00030000L

/* LABEL class type flags const */
#define LOCAL				0x01000000L
#define EXTLBL				0x02000000L
#define SECLBL				0x03000000L
#define GLBLBL				0x04000000L
#define COMMLBL				0x05000000L

/* MACRO class type flags */
#define DEFINED				0x01000000L

/* Macro to isolate the values for the type fields. */
#define IDType(x)			(x->Type & 0x00FF0000L)
#define IDFlag(x)			(x->Type & 0xFF000000L)

typedef struct symbol {
	char			*Name;
	long            Type;
	long			Value;

	long			offset;
	long			segment;

	struct patchref *backpatchlist; /* Backpatch list of the symbol */

	/* These pointers are used to keep track of symbols dependence */
	struct symbol	*NextLink;  /* Next symbol of the same dependence list */
	struct symbol	*Head;      /* Symbol which this one depends on. */

	/* These pointers are used to link the symbol in a 'unknown identifier'
	   list. Each symbol is removed from this list when it is defined. */
	struct symbol	*UnknownPrev, *UnknownNext;

	/* Extra informations */
	struct {
		long	__extra_long;
		void	*__extra_ptr;
	}			 __ExtraInfo;
} SYMBOL;

#define ExtraLong	__ExtraInfo.__extra_long
#define ExtraPtr	__ExtraInfo.__extra_ptr

extern SYMBOL	*LookUp(char *, int);
extern int		IsKeyword(char *);

extern void		FreeSymbolList(SYMBOL *);
extern void		LinkSymbol(SYMBOL *, SYMBOL *);
extern void		FreeSymbol(SYMBOL *);

#endif

/* --------- End of File --------------------------------------- */
